#include "mcdist.h"
#include "mcstats.h"
#include <fstream.h>
#include <iostream.h>


// filters out experiments according to some minimal perturbation criteria
// returns: a size of the filtered experiment vector
int filter_experiment_vectors(const vector<double>& pv1,const vector<double>& pv2,vector<int>& valid_experiments) {
  valid_experiments.clear();
  for(int i=0;i<pv1.size();i++) {
    if(pv1[i]<MIN_PVALUE_PERTURBATION && pv2[i]<MIN_PVALUE_PERTURBATION) {
      valid_experiments.push_back(i);
    }
  }
  return valid_experiments.size();
}



pair<double,double> calculate_PValue_distance(const string& gene1,const string& gene2,const GeneExpressionMapType& pvalue_expression_map,const GeneExpressionMapType&  ratio_expression_map,double& experiment_count_reduction,int& experiment_count) {
  GeneExpressionMapType::const_iterator i_gene1,i_gene2;
  
  if((i_gene1=pvalue_expression_map.find(gene1))!=pvalue_expression_map.end() && (i_gene2=pvalue_expression_map.find(gene2))!=pvalue_expression_map.end()) {
    const vector<double>& pvi=i_gene1->second;
    const vector<double>& pvj=i_gene2->second;
    if(pvi.size()!=pvj.size()) {
      cout<<"ERROR: size of experiment vectors for the following genes differ: "<<gene1<<" ("<<pvi.size()<<") and "<<gene2<<" ("<<pvj.size()<<")"<<endl<<flush;
      exit(1);
    }
    vector<int>* valid_experiments_ptr;
#ifdef OMIT_UNPERTURBED_EXPERIMENTS
    // filter out the expression vectors
    vector<int> valid_experiments(0);
    int n_valid_experiments=filter_experiment_vectors(pvi,pvj,valid_experiments);
    valid_experiments_ptr=&valid_experiments;
    // record size reduction
    experiment_count_reduction=(((double)experiment_count)*experiment_count_reduction+(((double)valid_experiments.size())/((double)pvi.size())))/((double)++experiment_count);
    //if(valid_experiments.size()!=pvi.size()) {
    //  cout<<gene1<<" : "<<gene2<<" nexp="<<valid_experiments.size()<<endl;
    //}
#else
    int n_valid_experiments=pvi.size();
    valid_experiments_ptr=0;
#endif
    if(n_valid_experiments>=MIN_VALID_EXPERIMENT_FRACTION*pvi.size()) {
      /*
      // make sure that at least one of the genes has been perturbed during the series of experiments
      if(min(*min_element(pvi.begin(),pvi.end()),*min_element(pvj.begin(),pvj.end()))>MIN_PVALUE_PERTURBATION) {
      //cout<<min(*min_element(pv1.begin(),pv1.end()),*min_element(pv2.begin(),pv2.end()))<<" ";
      dv.push_back(pair<double,double>(-1.0,0.0));
      } else {
      */
      switch (DISTANCE_MEASURE) {
      case SPEARMANK_RANK_CORRELATION_DISTANCE:
        if(DISTANCE_DATA==USE_PVALUE_DATA) {
          return calculate_PValue_spearman_distance(pvi,ratio_expression_map.find(gene1)->second,pvj,ratio_expression_map.find(gene2)->second,valid_experiments_ptr);
        } else if (DISTANCE_DATA==USE_RATIO_DATA) {
          return calculate_ratio_spearman_distance(ratio_expression_map.find(gene1)->second,ratio_expression_map.find(gene2)->second,valid_experiments_ptr);
        } else {
          cerr<<"mcdist.C: calculate_PValue_distance() : spearman correlation : unsupported data type "<<DISTANCE_DATA<<endl;
          exit(1);
        }
        break;
      case MUTUAL_INFORMATION_DISTANCE:
        if(DISTANCE_DATA==USE_PVALUE_DATA) {
        return calculate_PValue_mi_distance(pvi,ratio_expression_map.find(gene1)->second,pvj,ratio_expression_map.find(gene2)->second,valid_experiments_ptr);
        } else {
          cerr<<"mcdist.C: calculate_PValue_distance() : mutual information : unsupported data type "<<DISTANCE_DATA<<endl;
          exit(1);
        }
        break;
      case PEARSON_CORRELATION_DISTANCE:
        if(DISTANCE_DATA==USE_PVALUE_DATA) {
          return calculate_PValue_pearson_distance(pvi,ratio_expression_map.find(gene1)->second,pvj,ratio_expression_map.find(gene2)->second,valid_experiments_ptr);
        } else if (DISTANCE_DATA==USE_RATIO_DATA) {
          return calculate_ratio_pearson_distance(ratio_expression_map.find(gene1)->second,ratio_expression_map.find(gene2)->second,valid_experiments_ptr);
        } else {
          cerr<<"mcdist.C: calculate_PValue_distance() : pearson correlation : unsupported data type "<<DISTANCE_DATA<<endl;
          exit(1);
        }
        break;
      }
    } else {
      return pair<double,double>(-1.0,0.0);
    }
  } else {
     return pair<double,double>(-1.0,0.0);
  }
}
 
pair<double,double> calculate_PValue_distance(const string& gene1,const string& gene2,const GeneExpressionMapType& pvalue_expression_map,const GeneExpressionMapType&  ratio_expression_map) {
  double d_temp; int i_temp;
  return calculate_PValue_distance(gene1,gene2,pvalue_expression_map,ratio_expression_map,d_temp,i_temp);
}


// calculate distance matrix using PValues in expression data
void calculate_PValue_distance_matrix(vector< vector< pair<double,double> > >& distances,GeneExpressionMapType& pvalue_expression_map,GeneExpressionMapType&  ratio_expression_map,vector<string>& gene_names) {
  cout<<"calculating distances "<<flush;

  double experiment_count_reduction=0;
  int experiment_counts=0;

  for(vector<string>::iterator gi=gene_names.begin();gi!=gene_names.end();++gi) {
    cout<<"."<<flush;
    //    vector< pair<double,double> > dv(nGenes,pair<double,double>(0.0,0.0));

    // do we have expression information for this gene ?
    if(pvalue_expression_map.find(*gi)!=pvalue_expression_map.end()) {
      vector< pair<double,double> > dv;
      //      for(vector<string>::iterator gj=gene_names.begin();gj!=gene_names.end();++gj) {
      for(vector<string>::iterator gj=gene_names.begin();gj!=gi;++gj) {
        if(gi!=gj) {
          dv.push_back(calculate_PValue_distance(*gi,*gj,pvalue_expression_map,ratio_expression_map,experiment_count_reduction,experiment_counts));
        } else {
          dv.push_back(pair<double,double>(0.0,0.0));
        }
      }
      distances.push_back(dv);
    } else {
      
      distances.push_back( vector< pair<double,double> >(gene_names.size(),pair<double,double>(-1.0,0.0)));
    }
  }
  cout<<"done.\n"<<flush;
#ifdef OMIT_UNPERTURBED_EXPERIMENTS  
  cout<<"average experiment count reduction is "<<experiment_count_reduction<<endl<<flush;
#endif
}



void output_distance_vector(ostream& output,vector< pair<double,double> >& dv) {
  for(vector< pair<double,double> >::iterator pi=dv.begin();pi!=dv.end();++pi) {
    output<<pi->first<<' '<<pi->second<<'\t';
  }
}

// output distance matrix using PValues in expression data
void output_PValue_distance_matrix(string& output_fname,GeneExpressionMapType& pvalue_expression_map,GeneExpressionMapType&  ratio_expression_map,vector<string>& gene_names,bool output_verbose_file) {
  ofstream output(output_fname.c_str(),ios::binary | ios::out);
  //  fout<<setprecision(0);

  cout<<"calculating distance matrix of "<<gene_names.size()<<" genes"<<flush;

  int ngenes=gene_names.size();  
  float f_size_d_10=((float) ngenes)/10.0;

  double experiment_count_reduction=0;
  int experiment_counts=0;

  // output header line with gene names
  //  copy(gene_names.begin(),gene_names.end(),ostream_iterator<string>(output,"\t"));cout<<endl;
  copy(gene_names.begin(),gene_names.end(),ostream_iterator<string>(output,"\t"));output<<endl<<flush;

  size_t asize=ngenes*(ngenes-1)/2*sizeof(float);
  float *distances=(float*)malloc(asize);
  float *variances=(float*)malloc(asize);

  pair<double,double> dp_temp;
  int pos=0;

  int percent_count=0;
  for(vector<string>::iterator gi=gene_names.begin();gi!=gene_names.end();++gi) {
    cout<<'.'<<flush;
    int ipos=gi-gene_names.begin();
    float tfp=((float) (ipos))/f_size_d_10;
    int temp_percent=(int) (tfp*tfp);
    if(temp_percent>percent_count) {
      percent_count=temp_percent; cout<<percent_count<<'%'<<flush;
    }
    
    //    vector< pair<double,double> > dv(nGenes,pair<double,double>(0.0,0.0));

    // do we have expression information for this gene ?
    if(pvalue_expression_map.find(*gi)!=pvalue_expression_map.end()) {
      for(vector<string>::iterator gj=gene_names.begin();gj!=gi;++gj) {
        int jpos=gj-gene_names.begin();

        if(pos!=ipos*(ipos-1)/2+jpos) cout<<"ERROR: position mismatch pos="<<pos<<" cpos="<<ipos*(ipos-1)/2+jpos<<" ipos="<<ipos<<" jpos="<<jpos<<endl<<flush;
        
        dp_temp=calculate_PValue_distance(*gi,*gj,pvalue_expression_map,ratio_expression_map,experiment_count_reduction,experiment_counts);
        distances[pos]=(float)dp_temp.first;
        variances[pos]=(float)dp_temp.second;
        pos++;
      }
    } else {
      cout<<"ERROR: gene "<<*gi<<" is not in the expression map!!"<<endl<<flush;
    }
  }
  cout<<"done.\n"<<flush;

  // output arrays
  cout<<"writing out distances[] ... "<<flush;
  output.write((const char*)distances,asize);
  cout<<"done"<<endl<<"writing out variances[] ... "<<flush;
  output.write((const char*)variances,asize);
  cout<<"done"<<endl<<flush;
  output.close();

  if(output_verbose_file) {
    // output verbose table
    ofstream vout("all_distances.dat");
    vout<<"gene1"<<'\t'<<"gene2"<<'\t'<<"distance"<<'\t'<<"corr"<<endl;
    for(vector<string>::iterator gi=gene_names.begin();gi!=gene_names.end();++gi) {  
      int ipos=gi-gene_names.begin();
      if(pvalue_expression_map.find(*gi)!=pvalue_expression_map.end()) {
        for(vector<string>::iterator gj=gene_names.begin();gj!=gi;++gj) {
          if(pvalue_expression_map.find(*gj)!=pvalue_expression_map.end()) {
            int jpos=gj-gene_names.begin();
            int pos=ipos*(ipos-1)/2+jpos;
            vout<<*gi<<'\t'<<*gj<<'\t'<<distances[pos]<<'\t'<<variances[pos]<<endl;
          }
        }
      }
    }
    vout.close();
  }

  
  free(distances);
  free(variances);
#ifdef OMIT_UNPERTURBED_EXPERIMENTS  
  cout<<"average experiment count reduction is "<<experiment_count_reduction<<endl<<flush;
#endif
}

