#include "mcstats.h"
#include "boost/utility.hpp"
#include <gsl/gsl_cdf.h>
#include <algorithm>
#include <numeric>
#include <nr.h>

inline int discrete_value(double pv,double iv) {
  if(pv<PVALUE_CUTOFF) {
    if(iv<0) {
      return 0;
    } else {
      return 2;
    }
  } else {
    return 1;
  }
}

int fill_signed_pvalue_array(vector<double>& pv,vector<double>& rv,float f[],vector<int>* valid_experiments) {
  if(valid_experiments) {
    int nexp=valid_experiments->size();
    for(int i=0;i<nexp;++i) {
      if(rv[ (*valid_experiments)[i] ]<0) {
        f[i]=(float) (- pv[ (*valid_experiments)[i] ]);
      } else {
        f[i]=(float) (pv[ (*valid_experiments)[i] ]);
      }
    }
    return nexp;
  } else {
    int nexp=pv.size();
    for(int i=0;i<nexp;++i) {
      if(rv[i]<0) {
        f[i]=(float) (- pv[i]);
      } else {
        f[i]=(float) (pv[i]);
      }
    }
    return nexp;
  }
}

// generate a standard normal Z score (positive) from a p value
double z_score(double p_value) {
  double z= gsl_cdf_ugaussian_Qinv(p_value/2.0);
  // cout<<"z("<<p_value<<")="<<z<<endl;
  return z;
}

Vec_DP* generate_signed_pvalue_array(const vector<double>& pv,const vector<double>& rv,const vector<int>* valid_experiments) {
  if(valid_experiments) {
    int nexp=valid_experiments->size();
    Vec_DP* f=new Vec_DP(nexp);
    for(int i=0;i<nexp;++i) {
      if(rv[ (*valid_experiments)[i] ]<0) {
        (*f)[i]= (- z_score(pv[ (*valid_experiments)[i] ]));
      } else {
        (*f)[i]= (z_score(pv[ (*valid_experiments)[i] ]));
      }
    }
    return f;
  } else {
    int nexp=pv.size();
    Vec_DP* f=new Vec_DP(nexp);
    for(int i=0;i<nexp;++i) {
      if(rv[i]<0) {
        (*f)[i]= (- z_score(pv[i]));
      } else {
        (*f)[i]= (z_score(pv[i]));
      }
    }
    return f;
  }
}

Vec_DP* generate_ratio_array(const vector<double>& rv,const vector<int>* valid_experiments) {
  if(valid_experiments) {
    int nexp=valid_experiments->size();
    Vec_DP* f=new Vec_DP(nexp);
    for(int i=0;i<nexp;++i) {
      (*f)[i]=(rv[ (*valid_experiments)[i] ]);
    }
    return f;
  } else {
    int nexp=rv.size();
    Vec_DP* f=new Vec_DP(nexp);
    for(int i=0;i<nexp;++i) {
      (*f)[i]= (rv[i]);
    }
    return f;
  }
}

pair<double,double> calculate_PValue_spearman_distance(const vector<double>& pv1,const vector<double>& iv1,const vector<double>& pv2,const vector<double>& iv2,vector<int>* valid_experiments) {

  Vec_DP* nfc1=generate_signed_pvalue_array(pv1,iv1,valid_experiments);
  Vec_DP* nfc2=generate_signed_pvalue_array(pv2,iv2,valid_experiments);
  
  if(nfc1->size()<2 || nfc2->size()<2) {
    return pair<double,double>(-1.0,0.0);
  } else {
    double d,zd,probd,rs,probrs;
    NR::spear(*nfc1,*nfc2,d,zd,probd,rs,probrs);
    double r= (double) (1.0-fabs(rs));
    // calculate variance
    //double s=1.0/((double)(nfc1->size()-1));
    // The following line can be used to output signed correlation values
    double s=rs;
    delete nfc1; delete nfc2;
    return pair<double,double>(r,s);
  }
}

pair<double,double> calculate_ratio_spearman_distance(const vector<double>& rv1,const vector<double>& rv2,vector<int>* valid_experiments) {

  Vec_DP *nfc1=generate_ratio_array(rv1,valid_experiments);
  Vec_DP *nfc2=generate_ratio_array(rv2,valid_experiments);

  if(nfc1->size()<2 || nfc2->size()<2) {
    return pair<double,double>(-1.0,0.0);
  } else {
    double d,zd,probd,rs,probrs;
    NR::spear(*nfc1,*nfc2,d,zd,probd,rs,probrs);
    double r= (double) (1.0-fabs(rs));
    // calculate variance
    // double s=1.0/((double)(nfc1->size()-1));
    // The following line can be used to output signed correlation values
    double s=rs;
    delete nfc1; delete nfc2;
    return pair<double,double>(r,s);
  }
}


// calcualte distance between a pair of expression profiles using Spearman's rank correlation
pair<double,double> calculate_PValue_pearson_distance(const vector<double>& pv1,const vector<double>& iv1,const vector<double>& pv2,const vector<double>& iv2,vector<int>* valid_experiments) {
  Vec_DP* nfc1=generate_signed_pvalue_array(pv1,iv1,valid_experiments);
  Vec_DP* nfc2=generate_signed_pvalue_array(pv2,iv2,valid_experiments);

  if(nfc1->size()<2 || nfc2->size()<2) {
    return pair<double,double>(-1.0,0.0);
  } else {

    double r,p,z;
    NR::pearsn(*nfc1,*nfc2,r,p,z);


    double d= (double) (1.0-fabs(r));
    // calculate variance
    //double s=1.0/((double)(nfc1->size()-1));
    double s=r;
    delete nfc1; delete nfc2;
    return pair<double,double>(d,s);
  }
}


// calcualte distance between a pair of expression profiles using Spearman's rank correlation on ratio data
pair<double,double> calculate_ratio_pearson_distance(const vector<double>& rv1,const vector<double>& rv2,vector<int>* valid_experiments) {
  Vec_DP *nfc1=generate_ratio_array(rv1,valid_experiments);
  Vec_DP *nfc2=generate_ratio_array(rv2,valid_experiments);

  if(nfc1->size()<2 || nfc2->size()<2) {
    return pair<double,double>(-1.0,0.0);
  } else {

    double r,p,z;
    NR::pearsn(*nfc1,*nfc2,r,p,z);


    double d= (double) (1.0-fabs(r));
    // calculate variance
    // double s=1.0/((double)(nfc1->size()-1));
    // output correlation instead of variance
    double s=r;
    delete nfc1; delete nfc2;
    return pair<double,double>(d,s);
  }
}


// calculate distance between a pair of expression profiles
pair<double,double> calculate_PValue_mi_distance(const vector<double>& pv1,const vector<double>& iv1,const vector<double>& pv2,const vector<double>& iv2,vector<int>* valid_experiments) {

  int c1[3];
  int c2[3];
  int cm[3][3];
  memset(c1,0,3*sizeof(int));
  memset(c2,0,3*sizeof(int));
  memset(cm,0,9*sizeof(int));

  /*  
  vector<int> c1(3,0);
  vector<int> c2(3,0);
  vector< vector<int> > cm(3, vector<int>(3,0));
  */

  int nExp;
  if(valid_experiments) {
    nExp=valid_experiments->size();
    for(int i=0;i<nExp;i++) {
      int d1=discrete_value(pv1[(*valid_experiments)[i]],iv1[(*valid_experiments)[i]]);
      int d2=discrete_value(pv2[(*valid_experiments)[i]],iv2[(*valid_experiments)[i]]);
      c1[d1]++; c2[d2]++; cm[d1][d2]++;
    }
  } else {
    nExp=pv1.size();
    for(int i=0;i<nExp;i++) {
      int d1=discrete_value(pv1[i],iv1[i]);
      int d2=discrete_value(pv2[i],iv2[i]);
      c1[d1]++; c2[d2]++; cm[d1][d2]++;
    }
  }
  /*
  cout<<" c1:"<<c1<<endl;
  cout<<" c2:"<<c2<<endl;
  cout<<" cm:"<<cm[0]<<endl;
  cout<<" cm:"<<cm[1]<<endl;
  cout<<" cm:"<<cm[2]<<endl;
  */
  
  double h1=0;
  double h2=0;
  double mh=0;
  double d_nExp=((double) nExp);
  double d_nExp2=d_nExp*d_nExp;
  for(int i=0;i<3;i++) {
    if(c1[i]!=0) 
      h1-=((double) c1[i])/(d_nExp)*((double) log(((double) c1[i])/d_nExp));
    if(c2[i]!=0)  
      h2-=((double) c2[i])/(d_nExp)*((double) log(((double) c2[i])/d_nExp));
    for(int j=0;j<3;j++) {
      if(cm[i][j]!=0) 
        mh-=((double) cm[i][j])/(d_nExp)*((double) log(((double)cm[i][j])/d_nExp));
    }
  }
  double mi=h1+h2-mh;
  //  cout<<"h1="<<h1<<" h2="<<h2<<" mh="<<mh<<" mi="<<mi/log(2)<<endl;
  mi=mi/(h2+h1);
  /*
  if(h1<h2) {
    mi=mi/h2;
  } else {
    mi=mi/h1;
  }*/
  return pair<double,double>(1.0-mi,0.0);
}
