#include "mctypes.h"
#include "npdist.h"
#include "mcdist.h"
#include "mcio.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <strstream>
#include <boost/tokenizer.hpp>

using namespace std;

pair<double,double> NPCalcDist::get_distance(const string& gene1,const string& gene2) const {
  if(gene1!=gene2) {
    return calculate_PValue_distance(gene1,gene2,*pvalue_map,*ratio_map);
  } else {
    return pair<double,double>(0.0,0.0);
  }
};

vector<string> NPCalcDist::get_valid_genes() const {
  vector<string> v;
  for(GeneExpressionMapType::iterator i=pvalue_map->begin();i!=pvalue_map->end();++i) {
    v.push_back(i->first);
  }
  return v;
};

vector<string> NPArrayDist::get_valid_genes() const {
  return gene_names;
};

NPArrayDist::NPArrayDist(string& precalculated_distances_fname) {
  // open file
  ifstream distances_file(precalculated_distances_fname.c_str(),ios::binary | ios::in);
  // read the first line with the gene names
  string line;
  if(getline(distances_file,line)) {
    typedef boost::tokenizer<boost::char_separator<char> >  tokType;
    boost::char_separator<char> sep("\t");
    tokType tabtok(line, sep);
    copy(tabtok.begin(),tabtok.end(), back_inserter(gene_names));
    ngenes=gene_names.size();
    cout<<"reading precalculated distances for "<<ngenes<<" genes:"<<endl;
    
    // record gene ids
    ids.clear();
    for(vector<string>::iterator i=gene_names.begin();i!=gene_names.end();++i) {
      ids[*i]=i-gene_names.begin();
    }
      
    // allocate arrays
    int asize=(ngenes)*(ngenes-1)/2;
    //    cout<<"array size="<<asize<<'*'<<sizeof(float)<<'='<<asize*sizeof(float)<<endl;
    distances=(float*)malloc(asize*sizeof(float));
    variances=(float*)malloc(asize*sizeof(float));
    
    // fill in the array
    cout<<"reading in the distances data ... "<<flush;
    distances_file.read((char*)distances,asize*sizeof(float));
    cout<<"done"<<endl<<flush;

    cout<<"reading in the variances data ... "<<flush;
    distances_file.read((char*)variances,asize*sizeof(float));
    cout<<"done"<<endl<<flush;
    
  } else {
    cout<<"NPArrayDist::NPArrayDist() : ERROR: unable to read header line in the distances file "<<precalculated_distances_fname<<endl;
  }
};

int NPArrayDist::distance_position(int gene1, int gene2) const{
  if(gene1>gene2) {
    return ((gene1-1)*gene1/2)+gene2;
  } else {
    return ((gene2-1)*gene2/2)+gene1;
  }
};

pair<double,double> NPArrayDist::get_distance(const string& gene1,const string& gene2) const {
  // look up gene positions
  GeneIdMap::const_iterator gi1,gi2;
  gi1=ids.find(gene1); gi2=ids.find(gene2);
  if(gi1!=ids.end() && gi2!=ids.end()) {
    int id1=gi1->second;
    int id2=gi2->second;
    if(id1==id2) {
      return pair<double,double>(0.0,0.0);
    } else {
      int pos=distance_position(id1,id2);
      return pair<double,double>(((double)distances[pos]),((double)variances[pos]));
    }
  } else {
    return pair<double,double>(-1.0,0.0);
  }
};


NPConditionDist::NPConditionDist(string& pvalue_fname,string& ratio_fname,string& gdel_mapping_fname) : pvalue_map(), ratio_map(), gidm() {
  // read the gene deletion to experiment number mapping
  Properties<int,string> gdel_id_mapping(gdel_mapping_fname);
  cout<<"properties file:"<<endl;
  gdel_id_mapping.save(cout); 

  // read in the expression data
  GeneExpressionMapType raw_pvalue_map;
  GeneExpressionMapType raw_ratio_map;
  vector<string> experiment_names;
  GeneSetType gene_set;
  cout<<"reading in pvalue expression data from \""<<pvalue_fname<<"\" ... "<<flush;
  if(!read_expression(experiment_names,raw_pvalue_map,gene_set,pvalue_fname,0,false,true)) {
    cout<<"ERROR: unable to read pvalue expression data from \""+pvalue_fname+"\"";
  }
  cout<<"done"<<endl;
  
  
  // read in ratio data
  cout<<"reading in ratio expression data from \""<<ratio_fname<<"\" ... "<<flush;
  if(!read_expression(experiment_names,raw_ratio_map,gene_set,ratio_fname,0,false,true)) {
    cout<<"ERROR: unable to read ratio expression data from \""+ratio_fname+"\"";
  }
  cout<<"done"<<endl;
  
  // match up experiment names with experiment numbers
  cout<<"constructing condition expression vectors ... "<<flush;


  int i=0;
  for(GeneExpressionMapType::iterator ei=raw_pvalue_map.begin();ei!=raw_pvalue_map.end();++ei) {
    gidm[ei->first]=i++;
  }

  for(Properties<int,string>::iterator eii=gdel_id_mapping.begin();eii!=gdel_id_mapping.end();++eii) {
    int exp_num=eii->first-1;
    string gene_name=eii->second;
    cout<<"registering exp_num="<<exp_num<<" gene name="<<gene_name<<" exp_name="<<experiment_names[exp_num]<<endl<<flush;
    // map the gene name to the experiment number
    gdel_mapping.setProperty(gene_name,exp_num);
    // fill in the expression vector
    vector<double> condition_pvalues(0);
    for(GeneExpressionMapType::iterator ei=raw_pvalue_map.begin();ei!=raw_pvalue_map.end();++ei) {
      condition_pvalues.push_back((ei->second)[exp_num]);
    }
    pvalue_map[gene_name]=condition_pvalues;
    
    vector<double> condition_ratios(0);
    for(GeneExpressionMapType::iterator ei=raw_ratio_map.begin();ei!=raw_ratio_map.end();++ei) {
      condition_ratios.push_back((ei->second)[exp_num]);
    }
    ratio_map[gene_name]=condition_ratios;
    // remove this experiment -> gene mapping 
  }
  
  // check expression vector sizes
  int tsize=-1;
  for(GeneExpressionMapType::iterator mi=pvalue_map.begin();mi!=pvalue_map.end();++mi) {
    if(tsize==-1) {
      tsize=mi->second.size();
      cout<<"pvalue condition vector size = "<<tsize<<endl<<flush;
    } else {
      if(mi->second.size()!=tsize) {
        cout<<"ERROR: pvalue condition vector for condition "<<mi->first<<" has size "<<mi->second.size()<<" and not "<<tsize<<endl<<flush;
      }
    }
  }

  tsize=-1;
  for(GeneExpressionMapType::iterator mi=ratio_map.begin();mi!=ratio_map.end();++mi) {
    if(tsize==-1) {
      tsize=mi->second.size();
      cout<<"ratio condition vector size = "<<tsize<<endl<<flush;
    } else {
      if(mi->second.size()!=tsize) {
        cout<<"ERROR: ratio condition vector for condition "<<mi->first<<" has size "<<mi->second.size()<<" and not "<<tsize<<endl<<flush;
      }
    }
  }

  /*
  // print out a test experiment
  string sg("YAR015W");
  cout<<"names :[";
  for(GeneExpressionMapType::iterator ei=raw_ratio_map.begin();ei!=raw_ratio_map.end();++ei) {
    cout<<ei->first<<" ";
  } cout<<"]"<<endl;
  vector<double> pvv=pvalue_map[sg];
  cout<<sg<<" pv:["; copy(pvv.begin(),pvv.end(),ostream_iterator<double>(cout," "));cout<<"]"<<endl<<flush;
  vector<double> rvv=ratio_map[sg];
  cout<<sg<<" rv:["; copy(rvv.begin(),rvv.end(),ostream_iterator<double>(cout," "));cout<<"]"<<endl<<flush;

  */
  cout<<"done"<<endl;

  /*
  if(gdel_id_mapping.size()>0) {
    cout<<"the following condition mappings have not been found:"<<endl;
    gdel_id_mapping.save(cout); cout<<flush;
  } else {
    cout<<"all specified conditions have been found"<<endl<<flush;
  }
  */
};

vector<string> NPConditionDist::get_valid_genes() const {
  vector<string> v;
  for(GeneExpressionMapType::const_iterator i=pvalue_map.begin();i!=pvalue_map.end();++i) {
    v.push_back(i->first);
  }
  return v;
};

double NPConditionDist::get_pvalue(const string& condition_name,const string& gene_name) const {
  GeneIdMapType::const_iterator gi=gidm.find(gene_name);
  GeneExpressionMapType::const_iterator ci=pvalue_map.find(condition_name);
  if(ci==pvalue_map.end() || gi==gidm.end()) {
    return -1.0;
  } else {
    return (ci->second)[gi->second];
  }
}

pair<double,double> NPConditionDist::get_distance(const string& gene1,const string& gene2) const {
  if(gdel_mapping.containsKey(gene1) && gdel_mapping.containsKey(gene2)) {
    return calculate_PValue_distance(gene1,gene2,pvalue_map,ratio_map);
  } else {
    cout<<"NPConditionDist::get_distance() : invalid gene pair (\""<<gene1<<"\",\""<<gene2<<"\")"<<endl<<flush;
    return pair<double,double>(-1.0,0.0);
  }
};

// calculate distance matrix using PValues in expression data
void calculate_PValue_distance_matrix(vector< vector< pair<double,double> > >& distances,NPDist& npd,vector<string>& gene_names) {
  cout<<"calculating distances "<<flush;
  for(vector<string>::const_iterator gi=gene_names.begin();gi!=gene_names.end();++gi) {
    cout<<"."<<flush;
    //    vector< pair<double,double> > dv(nGenes,pair<double,double>(0.0,0.0));
    
    vector< pair<double,double> > dv;
    for(vector<string>::const_iterator gj=gene_names.begin();gj!=gi;++gj) {
      if(gi!=gj) {
        dv.push_back(npd.get_distance(*gi,*gj));
      } else {
        dv.push_back(pair<double,double>(0.0,0.0));
      }
    }
    distances.push_back(dv);
  }
  cout<<"done.\n"<<flush;
};



NPShuffledArrayDist::NPShuffledArrayDist(string& precalculated_distances_fname) : NPArrayDist(precalculated_distances_fname), shuffled_gene_name() {
  //  shuffle_genes();
  // use identity map as a starting point
  vector<string> shuffled_genes=gene_names;
  for(vector<string>::iterator i=shuffled_genes.begin();i!=shuffled_genes.end();++i) {
    shuffled_gene_name[gene_names[i-shuffled_genes.begin()]]=*i;
  }
}

vector<string> NPShuffledArrayDist::get_valid_genes() const {
  return NPArrayDist::get_valid_genes();
}

void NPShuffledArrayDist::shuffle_genes() {
  // clear the old mapping
  shuffled_gene_name.clear();
  vector<string> shuffled_genes=gene_names;
  random_shuffle(shuffled_genes.begin(),shuffled_genes.end());
  for(vector<string>::iterator i=shuffled_genes.begin();i!=shuffled_genes.end();++i) {
    shuffled_gene_name[gene_names[i-shuffled_genes.begin()]]=*i;
  }
}

pair<double,double> NPShuffledArrayDist::get_distance(const string& gene1,const string& gene2) const {
  GeneNameMapType::const_iterator ni1,ni2;
  ni1=shuffled_gene_name.find(gene1);
  ni2=shuffled_gene_name.find(gene2);
  if(ni1!=shuffled_gene_name.end() && ni2!=shuffled_gene_name.end()) {
    return NPArrayDist::get_distance(ni1->second,ni2->second);
  } else {
    // cerr<<"NPShuffledArrayDist::get_distance() : unable to find a map entry for genes \""<<gene1<<"\" or \""<<gene2<<"\""<<endl;
    return pair<double,double>(-1.0,0.0);
  }
};
