#include "npeval.h"
#include <math.h>
#include <algorithm>
#include <string>
#include <numeric>
#include <boost/iterator_adaptors.hpp>
#include <boost/function.hpp>
#include <boost/functional.hpp>
#include <boost/bind.hpp>
#include <boost/compose.hpp>
#include "mcdist.h"

using namespace std;

// define + operator on the pair<double,double>
namespace std {
  inline pair<double,double> operator+(const pair<double,double>& __x, const pair<double, double>& __y) {
    return pair<double,double>(__x.first+__y.first,__x.second +__y.second);
  };
};

// distinguish invalid distance pairs (those with negative distance values)
struct valid_distance_pair {
  bool operator()(const pair<double,double>& p) { return (p.first>=0.0); }
};

/** 
 * cost function for the gene placement algorithm
 * 
 * @param neighbour_distances  a vector (for neighbours 1,2,3.. mdistance away)
 *           where each vector contains a dist,variance pair for the set of
 *           genes neighbouring at that mdistance
 * 
 * @return a value and variance for this cost function
 */ 
pair<double,double> placement_cost_function(const vector< vector< pair<double,double> > >& neighbour_distances,const vector<double>& weights) {
  if(neighbour_distances.size()>weights.size()) {
    cout<<"ERROR: placement_cost_function() number of weight factors is less than the neighbourhood radius."<<endl<<flush;
    return pair<double,double>(-1.0,0.0);
  }
  
  // calculate mean for each neighbourhood radius
  pair<double,double> cost(0.0,0.0);

  for(vector< vector< pair<double,double> > >::const_iterator ni=neighbour_distances.begin();ni!=neighbour_distances.end();++ni) {
    if(ni->size()>0) {
      pair<double,double> ncost(0.0,0.0);
      vector< pair<double,double> > cv(boost::make_filter_iterator<valid_distance_pair>(ni->begin(), ni->end()),boost::make_filter_iterator<valid_distance_pair>(ni->end(), ni->end()));
      if(cv.size()>0) {
        ncost=accumulate(cv.begin(),cv.end(),ncost);
        double mult=weights[ni-neighbour_distances.begin()]/((double) cv.size());
        //  cout<<"nradius="<<ni-neighbour_distances.begin()<<" nsize="<<cv.size()<<" mult="<<mult<<" accum=("<<ncost.first<<","<<ncost.second<<")";
        ncost.first*=mult;
        ncost.second*=mult*mult;

        //ncost.second=1.0; // debug only

        cost=cost+ncost;
        //  cout<<" ncost=("<<ncost.first<<","<<ncost.second<<")";
        //  cout<<" cost=("<<cost.first<<","<<cost.second<<")"<<endl;

      }
    }
  }

  return cost;
}


/** 
 * cost function for the gene placement algorithm
 * 
 * @param neighbour_distances  a vector (for neighbours 1,2,3.. mdistance away)
 *           where each vector contains a dist,variance pair for the set of
 *           genes neighbouring at that mdistance
 * 
 * @return a value and variance for this cost function
 */ 
pair<double,double> placement_flat_cost_function(const vector< vector< pair<double,double> > >& neighbour_distances,const vector<double>& weights) {
  if(neighbour_distances.size()>weights.size()) {
    cout<<"ERROR: placement_flat_cost_function() number of weight factors is less than the neighbourhood radius."<<endl<<flush;
    return pair<double,double>(-1.0,0.0);
  }
  
  // calculate mean for each neighbourhood radius
  pair<double,double> cost(0.0,0.0);

  vector< pair<double,double> > all_costs;
  for(vector< vector< pair<double,double> > >::const_iterator ni=neighbour_distances.begin();ni!=neighbour_distances.end();++ni) {
    if(ni->size()>0) {
      all_costs.insert(all_costs.end(),boost::make_filter_iterator<valid_distance_pair>(ni->begin(), ni->end()),boost::make_filter_iterator<valid_distance_pair>(ni->end(), ni->end()));
    }
  }

  if(all_costs.size()>0) {
    cost=accumulate(all_costs.begin(),all_costs.end(),cost);
    double mult=1.0/((double)all_costs.size());
    cost.first*=mult; cost.second=mult*mult;
  } else {
    return pair<double,double>(-1.0,0.0);
  }

  return cost;
}

/** 
 * fitness function for the gene placement algorithm, that uses 1/distance
 * 
 * @param neighbour_distances  a vector (for neighbours 1,2,3.. mdistance away)
 *           where each vector contains a dist,variance pair for the set of
 *           genes neighbouring at that mdistance
 * 
 * @return a value and variance for this cost function
 */ 
pair<double,double> placement_fitness_function(const vector< vector< pair<double,double> > >& neighbour_distances,const vector<double>& weights) {
  if(neighbour_distances.size()>weights.size()) {
    cout<<"ERROR: placement_fitness_function() number of weight factors is less than the neighbourhood radius."<<endl<<flush;
    return pair<double,double>(-1.0,0.0);
  }
  
  // calculate mean for each neighbourhood radius
  pair<double,double> cost(0.0,0.0);

  for(vector< vector< pair<double,double> > >::const_iterator ni=neighbour_distances.begin();ni!=neighbour_distances.end();++ni) {
    if(ni->size()>0) {
      pair<double,double> ncost(0.0,0.0);
      vector< pair<double,double> > cv;
      valid_distance_pair vdp;
      for(vector< pair<double,double> >::const_iterator i=ni->begin();i!=ni->end();++i) {
        if(vdp(*i)) {
          cv.push_back(pair<double,double>(1.0/pow(i->first,POWER_FACTOR),i->second/(pow(i->first,4.0))));
          //cv.push_back(pair<double,double>(1.0/i->first,i->second/(pow(i->first,4.0))));
        }
      }

      if(cv.size()>0) {
        ncost=accumulate(cv.begin(),cv.end(),ncost);
        double mult=weights[ni-neighbour_distances.begin()]/((double) cv.size());
        //  cout<<"nradius="<<ni-neighbour_distances.begin()<<" nsize="<<cv.size()<<" mult="<<mult<<" accum=("<<ncost.first<<","<<ncost.second<<")";
        ncost.first*=mult;
        ncost.second*=mult*mult;

        //ncost.second=1.0; // debug only

        cost=cost+ncost;
        //  cout<<" ncost=("<<ncost.first<<","<<ncost.second<<")";
        //  cout<<" cost=("<<cost.first<<","<<cost.second<<")"<<endl;

      }
    }
  }
  return cost;
}


/** 
 * fitness function for the gene placement algorithm, that uses 1/distance^n and weights for the neighborhood size
 * 
 * @param neighbour_distances  a vector (for neighbours 1,2,3.. mdistance away)
 *           where each vector contains a dist,variance pair for the set of
 *           genes neighbouring at that mdistance
 * 
 * @return a value and variance for this cost function
 */ 
pair<double,double> placement_size_balanced_fitness_function(const vector< vector< pair<double,double> > >& neighbour_distances,const vector<double>& weights) {
  if(neighbour_distances.size()>weights.size()) {
    cout<<"ERROR: placement_fitness_function() number of weight factors is less than the neighbourhood radius."<<endl<<flush;
    return pair<double,double>(-1.0,0.0);
  }
  
  // calculate mean for each neighbourhood radius
  vector< pair<double,double> > cost_vector;
  pair<double,double> cost(0.0,0.0);
  int tsize=0;
  for(vector< vector< pair<double,double> > >::const_iterator ni=neighbour_distances.begin();ni!=neighbour_distances.end();++ni) {
    if(ni->size()>0) {
      pair<double,double> ncost(0.0,0.0);
      vector< pair<double,double> > cv;
      valid_distance_pair vdp;
      for(vector< pair<double,double> >::const_iterator i=ni->begin();i!=ni->end();++i) {
        if(vdp(*i)) {
          if(i->second>0.0) { // use only positively correlated pairs
            cv.push_back(pair<double,double>(1.0/pow(i->first,POWER_FACTOR),i->second/(pow(i->first,4.0))));
            // cv.push_back(pair<double,double>(1.0/pow(i->first,2.0),i->second/(pow(i->first,4.0))));
          }
        }
      }

      if(cv.size()>0) {
        ncost=accumulate(cv.begin(),cv.end(),ncost);
        double mult=weights[ni-neighbour_distances.begin()];
        ncost.first*=mult;
        ncost.second*=mult*mult;

        cost.first+=ncost.first*mult;
        cost.second+=ncost.second*mult*mult;
        tsize+=cv.size();
      }
    }
  }

  double sf=1.0/((double)tsize);
  cost.first*=sf; cost.second*=sf*sf;
  return cost;
}


/** 
 * fitness function for the gene placement algorithm, that uses 1/distance
 * 
 * @param neighbour_distances  a vector (for neighbours 1,2,3.. mdistance away)
 *           where each vector contains a dist,variance pair for the set of
 *           genes neighbouring at that mdistance
 * 
 * @return a value and variance for this cost function
 */ 
pair<double,double> placement_flat_fitness_function(const vector< vector< pair<double,double> > >& neighbour_distances,const vector<double>& weights) {
  if(neighbour_distances.size()>weights.size()) {
    cout<<"ERROR: placement_flat_fitness_function() number of weight factors is less than the neighbourhood radius."<<endl<<flush;
    return pair<double,double>(-1.0,0.0);
  }
  
  // calculate mean for each neighbourhood radius
  pair<double,double> cost(0.0,0.0);

  vector< pair<double,double> > all_costs;
  for(vector< vector< pair<double,double> > >::const_iterator ni=neighbour_distances.begin();ni!=neighbour_distances.end();++ni) {
    if(ni->size()>0) {
      all_costs.insert(all_costs.end(),boost::make_filter_iterator<valid_distance_pair>(ni->begin(), ni->end()),boost::make_filter_iterator<valid_distance_pair>(ni->end(), ni->end()));
    }
  }
  if(all_costs.size()>0) {
    for(vector< pair<double,double> >::const_iterator i=all_costs.begin();i!=all_costs.end();++i) {
        cost.first+=1.0/pow(i->first,12.0);
        cost.second+=i->second/(pow(i->first,4.0));
    }
    double mult=1.0/((double)all_costs.size());
    cost.first*=mult; cost.second=mult*mult;
  } else {
    return pair<double,double>(-1.0,0.0);
  }

  return cost;
}

/**
 * A utility method for computing expression distances between a gene and a given list of genes
 * using caclucate_PValue_distances method
 *
 * @param gene gene name
 * @param genes a vector of gene names
 * @param pvalue_exression_map
 * @param ratio_expression_map
 *
 * @return a vector of distances (value,variance)
 */
vector< pair<double,double> > compute_distance_to_gene(const string& gene,const vector<string>& genes,NPDist& npd) {

  typedef boost::function< pair<double,double> (const string&) > DistanceFunctionType;
  typedef boost::transform_iterator_generator<DistanceFunctionType, vector<string>::const_iterator>::type distance_comp_iterator;
   
  //  DistanceFunctionType name2distance=boost::bind(calculate_PValue_distance,boost::cref(gene),_1,boost::ref(pvalue_expression_map),boost::ref(ratio_expression_map));
  DistanceFunctionType name2distance=boost::bind(&NPDist::get_distance,&npd,boost::cref(gene),_1);
  distance_comp_iterator dc_begin(genes.begin(), name2distance);
  distance_comp_iterator dc_end(genes.end(), name2distance);

  return vector< pair<double,double> >(dc_begin,dc_end);
}

/**
 * A utility method for computing a set of distance vectors from a a given gene to a set of gene sets.
 * @param gene gene name
 * @param genes a vector of gene name vectors
 * @param pvalue_exression_map
 * @param ratio_expression_map
 * 
 * @return a vector of vector of distances (value,variance)
 */
vector< vector< pair<double,double> > > compute_distances_to_gene(const string& gene,const vector< vector<string> >& genes,NPDist& npd) {

  typedef boost::function< vector< pair<double,double> > (const vector<string>&) > DistancesFunctionType;
  typedef boost::transform_iterator_generator<DistancesFunctionType, vector< vector<string> >::const_iterator>::type distance_comp_iterator;
  
  DistancesFunctionType name2distance=boost::bind(compute_distance_to_gene,boost::cref(gene),_1,boost::ref(npd));
  distance_comp_iterator dc_begin(genes.begin(), name2distance);
  distance_comp_iterator dc_end(genes.end(), name2distance);
  return vector< vector< pair<double,double> > >(dc_begin,dc_end);
}

/**
 * Calculate placement scores for a set of genes with respect to a given position in the metabolic map.
 *
 * @param positionGene gene name defining a position in the metabolic map
 * @param genes a vector of genes to calculate the placement scores for
 * @param pvalue_exression_map
 * @param ratio_expression_map
 * @param npc NPConnectivity object describing the metabolic connectivity
 * @param weights a vector of weights to be used in calculating the cost function.
 *         Note: the size of the vector defines the neighbourhood radius
 * 
 * @return a vector of placement scores (value, variance)
 */
vector< pair<double,double> > score_genes_for_position(const string& positionGene, const vector<string>& genes,NPDist& npd,NPConnectivity& npc,const vector<double>& weights) {
  // determine genes neighbouring the specified position
  vector< vector<string> > neighbours=npc.get_neighbours(positionGene,weights.size());

  // check neighbours validity
  bool neighbourhood_valid=false;
  for(vector< vector<string> >::iterator ni=neighbours.begin();ni!=neighbours.end();++ni) {
    if(ni->size()>0) { neighbourhood_valid=true; break; }
  }
  if(!neighbourhood_valid) return vector< pair<double,double> >(0); // return invalid vector

  // define distance, cost and evaluation functional types
  typedef boost::function< pair<double,double> (const string&) > EvalFunction;

  // deine evaluation transformation iterator
  typedef boost::transform_iterator_generator<EvalFunction, vector<string>::const_iterator>::type cost_comp_iterator;

  // define evaluation function
  EvalFunction geneeval=boost::bind(placement_size_balanced_fitness_function,boost::bind(compute_distances_to_gene,_1,boost::cref(neighbours),boost::ref(npd)),boost::cref(weights));
 
  // define transformation endpoints
  cost_comp_iterator cc_begin(genes.begin(),geneeval);
  cost_comp_iterator cc_end(genes.end(),geneeval);
  // energize :)
  return vector< pair<double,double> >(cc_begin,cc_end);
}

/**
 * Determine the rank of a given metabolic gene in the nme_gene_set ordered by 
 * the cost function with respect to the location of that metabolic gene.
 */
pair<int,pair<double,double> > rank_gene(const string& metabolic_gene,const vector<string>& gene_vector,NPDist& npd,NPConnectivity& npc,const vector<double>& weights) {
  vector<string> genes(gene_vector);
  genes.push_back(metabolic_gene);
  // obtain gene scores
  vector< pair<double,double> > scores=score_genes_for_position(metabolic_gene,genes,npd,npc,weights);

  if(scores.size()!=genes.size()) {
    if(scores.size()!=0) {
      cout<<"rank_gene(): ERROR: score_genes_for_position() returned a score vector of length "<<scores.size()<<" for a gene vector of length "<<genes.size()<<endl<<flush;
    }
    return pair<int,pair<double,double> >(-1,pair<double,double>(0.0,0.0));
  }

  // determine the rank of the given metabolic gene
  pair<double,double> mg=scores[genes.size()-1];

  int rank=0;
  double min_dist=mg.first;
  for(vector< pair<double,double> >::iterator si=scores.begin();si!=scores.end();++si) {
    if(si->first >= mg.first) { 
      rank++; 
      if(si->first>min_dist) min_dist=si->first;
    }
    /*
    if(si->second!=mg.second) { // debug only
      cout<<" "<<mg.second<<"!=cvsize("<<genes[si-scores.begin()]<<")="<<si->second;
    }
    */
  }
  // cout<<"(mind="<<min_dist<<")";
  return pair<int,pair<double,double> >(rank,mg);
}

double self_rank_fitness(const gsl_vector *v, void *params) {
  // reconstruct the weights vector
  vector<double> weights;

  double pie=1.0;

  // pre-defined first weight
  //  double initial_weight=0.716;
  //  weights.push_back(initial_weight);
  //  double pie=1.0-initial_weight;

  for(int i=0;i<v->size;i++) {
    double ratio=gsl_vector_get(v,i);
    if(ratio >1.0 || ratio < 0.0) {
      return LARGE_COST;
    }

    // determine the slice
    double slice=pie*ratio;
    weights.push_back(slice);
    pie=pie-slice;
  }

  weights.push_back(pie); // last weight

  // reconstruct controls
  self_rank_fitness_controls *controls = ((self_rank_fitness_controls *)params);

  // self-rank all given metabolic genes
  double rank_sum=0;
  for(vector<string>::iterator mgi=controls->ptr_metabolic_genes->begin();mgi!=controls->ptr_metabolic_genes->end();++mgi) {
    rank_sum+=(double)rank_gene(*mgi,*(controls->ptr_nme_genes),*(controls->ptr_npd),*(controls->ptr_npc),weights).first;
  }
  return ((double)rank_sum)/((double)controls->ptr_metabolic_genes->size());
}


double self_logrank_fitness(const gsl_vector *v, void *params) {
  // reconstruct the weights vector
  vector<double> weights;

  double pie=1.0;

  // pre-defined first weight
  //  double initial_weight=0.716;
  //  weights.push_back(initial_weight);
  //  double pie=1.0-initial_weight;

  for(int i=0;i<v->size;i++) {
    double ratio=gsl_vector_get(v,i);
    if(ratio >1.0 || ratio < 0.0) {
      return LARGE_COST;
    }

    // determine the slice
    double slice=pie*ratio;
    weights.push_back(slice);
    pie=pie-slice;
  }

  weights.push_back(pie); // last weight

  // reconstruct controls
  self_rank_fitness_controls *controls = ((self_rank_fitness_controls *)params);

  // self-rank all given metabolic genes
  double log10_rank_sum=0;
  for(vector<string>::iterator mgi=controls->ptr_metabolic_genes->begin();mgi!=controls->ptr_metabolic_genes->end();++mgi) {
    int rank=rank_gene(*mgi,*(controls->ptr_nme_genes),*(controls->ptr_npd),*(controls->ptr_npc),weights).first;
    if(rank>0) {
      log10_rank_sum+=log10((double)rank);
    }
  }
  return ((double)log10_rank_sum)/((double)controls->ptr_metabolic_genes->size());
}



double self_max50_fitness(const gsl_vector *v, void *params) {
  // reconstruct the weights vector
  vector<double> weights;

  double pie=1.0;

  // pre-defined first weight
  //  double initial_weight=0.716;
  //  weights.push_back(initial_weight);
  //  double pie=1.0-initial_weight;

  for(int i=0;i<v->size;i++) {
    double ratio=gsl_vector_get(v,i);
    if(ratio >1.0 || ratio < 0.0) {
      return LARGE_COST;
    }

    // determine the slice
    double slice=pie*ratio;
    weights.push_back(slice);
    pie=pie-slice;
  }

  weights.push_back(pie); // last weight

  // reconstruct controls
  self_rank_fitness_controls *controls = ((self_rank_fitness_controls *)params);

  // self-rank all given metabolic genes
  double count=0;
  for(vector<string>::iterator mgi=controls->ptr_metabolic_genes->begin();mgi!=controls->ptr_metabolic_genes->end();++mgi) {
    int rank=rank_gene(*mgi,*(controls->ptr_nme_genes),*(controls->ptr_npd),*(controls->ptr_npc),weights).first;
    if(rank>0 && rank <=50) {
      count++;
    }
  }
  return 1.0/((double)count);
}

double self_psum_fitness(const gsl_vector *v, void *params) {
  // reconstruct the weights vector
  vector<double> weights;

  double pie=1.0;

  // pre-defined first weight
  //  double initial_weight=0.716;
  //  weights.push_back(initial_weight);
  //  double pie=1.0-initial_weight;

  for(int i=0;i<v->size;i++) {
    double ratio=gsl_vector_get(v,i);
    if(ratio >1.0 || ratio < 0.0) {
      return LARGE_COST;
    }

    // determine the slice
    double slice=pie*ratio;
    weights.push_back(slice);
    pie=pie-slice;
  }

  weights.push_back(pie); // last weight

  // reconstruct controls
  self_rank_fitness_controls *controls = ((self_rank_fitness_controls *)params);

  // self-rank all given metabolic genes
  double psum=0.0;
  for(vector<string>::iterator mgi=controls->ptr_metabolic_genes->begin();mgi!=controls->ptr_metabolic_genes->end();++mgi) {
    int rank=rank_gene(*mgi,*(controls->ptr_nme_genes),*(controls->ptr_npd),*(controls->ptr_npc),weights).first;

    if(rank>0) {
      psum+=1/((double)rank);
    }
  }
  return psum;
}





/**
 * Measure neighborhood variability
 *
 * @param positionGene gene name defining a position in the metabolic map
 * @param npc NPConnectivity object describing the metabolic connectivity
 * @param pvalue_exression_map
 * @param weights a vector of weights to be used in calculating the cost function.
 *         Note: the size of the vector defines the neighbourhood radius
 * 
 * @return a a pair containing a pair of neighborhood and self-variability measures and the vector of vectors of the neighborhood genes
 */
pair< pair<double, double>, vector< vector<string> > > calc_neighborhood_variability(const string& positionGene,NPConnectivity& npc,const GeneExpressionMapType& pvalue_expression_map,const vector<double>& weights) {
  double self_variability=-1.0; // default value
  GeneExpressionMapType::const_iterator i_gene=pvalue_expression_map.find(positionGene);
  if(i_gene!=pvalue_expression_map.end()) {
    vector<int> selected_experiments;
    self_variability=filter_experiment_vectors(i_gene->second,i_gene->second,selected_experiments);    
  }

  // determine genes neighbouring the specified position
  vector< vector<string> > neighbours=npc.get_neighbours(positionGene,weights.size());

  vector< vector<double> > neighborhood_variability;
  for(vector< vector<string> >::iterator ni=neighbours.begin();ni!=neighbours.end();++ni) {
    vector<double> nv;
    for(vector<string>::iterator pi=ni->begin();pi!=ni->end();++pi) {
      vector<int> selected_experiments;
      GeneExpressionMapType::const_iterator i_gene=pvalue_expression_map.find(*pi);
      if(i_gene!=pvalue_expression_map.end()) {
        nv.push_back(filter_experiment_vectors(i_gene->second,i_gene->second,selected_experiments));
      }
    }
    neighborhood_variability.push_back(nv);
  }

  // calculate the means
  /*
  // mean of means method
  vector<double> nmeans;
  for(vector< vector<double> >::iterator i=neighborhood_variability.begin();i!=neighborhood_variability.end();++i) {
    if(i->size()>0) {
      nmeans.push_back(accumulate(i->begin(),i->end(),0.0)/i->size());
    }
  }
  */

  // mean method
  vector<double> nmeans;
  for(vector< vector<double> >::iterator i=neighborhood_variability.begin();i!=neighborhood_variability.end();++i) {
    //      cout<<"nmeans=[";
    //      copy(nmeans.begin(),nmeans.end(),ostream_iterator<double>(cout," "));
    //      cout<<"]"<<endl<<flush;
    // nmeans.insert(nmeans.end(),i->begin(),i->end());
    nmeans.resize(nmeans.size()+i->size());
    transform(i->begin(),i->end(),nmeans.end()-i->size(),bind2nd(multiplies<double>(),weights[i-neighborhood_variability.begin()]/((double)i->size())));
    // transform(i->begin(),i->end(),nmeans.end()-i->size(),bind2nd(multiplies<double>(),weights[i-neighborhood_variability.begin()]));
  }
  
  pair<double,double> vp=pair<double,double>(-1.0,self_variability);
  if(nmeans.size()>0) { 
    vp.first=accumulate(nmeans.begin(),nmeans.end(),0.0);
    // vp.first=accumulate(nmeans.begin(),nmeans.end(),0.0)/nmeans.size();
  }

  return pair< pair<double, double>, vector< vector<string> > >(vp,neighbours);
}


/**
 * Measure neighborhood variability vector.
 */
/*
pair< vector< vector<double> >, vector< vector<string> > > calc_neighborhood_variability_vector(const string& positionGene,NPConnectivity& npc,const GeneExpressionMapType& pvalue_expression_map,const int radius) {
  // determine genes neighbouring the specified position
  vector< vector<string> > neighbours=npc.get_neighbours(positionGene,radius);
  vector< vector<double> > neighborhood_variability;

  vector<int> selected_experiments;
  
  // add the place itself to the list
  {
    vector<double> nv0;
    GeneExpressionMapType::const_iterator z_gene=pvalue_expression_map.find(positionGene);
    if(z_gene!=pvalue_expression_map.end()) {
      nv0.push_back(filter_experiment_vectors(z_gene->second,z_gene->second,selected_experiments));
    }
    neighborhood_variability.push_back(nv0);
  }

  for(vector< vector<string> >::iterator ni=neighbours.begin();ni!=neighbours.end();++ni) {
    vector<double> nv;
    for(vector<string>::iterator pi=ni->begin();pi!=ni->end();++pi) {
      GeneExpressionMapType::const_iterator i_gene=pvalue_expression_map.find(*pi);
      if(i_gene!=pvalue_expression_map.end()) {
        nv.push_back(filter_experiment_vectors(i_gene->second,i_gene->second,selected_experiments));
      }
    }
    neighborhood_variability.push_back(nv);
  }

  return pair< vector< vector<double> >, vector< vector<string> > >(neighborhood_variability,neighbours);
}
*/
